/*
 * Copyright (C) 2004,2005 Charles Schmidt <cschmidt2@emich.edu>
 * Copyright (C) 2006 INDT
 *  Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "daap-connection.h"
#include "daap-root-item.h"
#include "dmap-md5.h"
#include "dmap-structure.h"

#include <string.h>

#include <libsoup/soup.h>

#define DAAP_USER_AGENT "iTunes/4.6 (Macintosh; PPC)"

static SoupMessage *daap_connection_build_message (DMAPConnection *self,
                                                   SoupURI *base_uri,
                                                   const gchar *path,
                                                   gboolean need_hash,
                                                   gdouble version,
                                                   gint req_id);
static DMAPItem *daap_connection_create_root_item (DMAPConnection *self);

G_DEFINE_TYPE (DAAPConnection, daap_connection, TYPE_DMAP_CONNECTION)

static void
daap_connection_class_init (DAAPConnectionClass *klass)
{
    DMAPConnectionClass *dmap_connection_class = (DMAPConnectionClass *) klass;

    dmap_connection_class->build_message = daap_connection_build_message;
    dmap_connection_class->create_root_item = daap_connection_create_root_item;
}

static void
daap_connection_init (DAAPConnection *self)
{
}

DAAPConnection *
daap_connection_new (const gchar *name,
                     const gchar *host,
                     gint port,
                     const gchar *password)
{
    DAAPConnection *self = DAAP_CONNECTION (g_object_new (TYPE_DAAP_CONNECTION,
            "name", name,
            "host", host,
            "port", port,
            "need-revision-number", TRUE,
            NULL));

    if (password) {
        g_object_set (self, "password", password, NULL);
    }

    return self;
}

static SoupMessage *
daap_connection_build_message (DMAPConnection *self,
                               SoupURI *base_uri,
                               const gchar *path,
                               gboolean need_hash,
                               gdouble version,
                               gint req_id)
{
    SoupMessage *message = NULL;
    SoupURI *uri = NULL;

    uri = soup_uri_new_with_base (base_uri, path);
    if (uri == NULL) {
        return NULL;
    }

    message = soup_message_new_from_uri (SOUP_METHOD_GET, uri);
    soup_message_set_http_version (message, SOUP_HTTP_1_1);

    soup_message_headers_append (message->request_headers, "Client-DAAP-Version",   "3.0");
    soup_message_headers_append (message->request_headers, "Accept-Language",       "en-us, en;q=5.0");
#ifdef HAVE_LIBZ
    soup_message_headers_append (message->request_headers, "Accept-Encoding",       "gzip");
#endif
    soup_message_headers_append (message->request_headers, "Client-DAAP-Access-Index",  "2");

    if (need_hash) {
        gchar hash[33] = {0};
        dmap_hash_generate ((short) floor (version), (const guchar*) path, 2, (guchar*) hash, req_id);
        soup_message_headers_append (message->request_headers, "Client-DAAP-Validation", hash);
    }

    soup_uri_free (uri);

    return message;
}

static DMAPItem *
daap_connection_create_root_item (DMAPConnection *self)
{
    DMAPItem *item;
    gchar *name;
    g_object_get (G_OBJECT (self),
            "name", &name, NULL);
    item = DMAP_ITEM (g_object_new (TYPE_DAAP_ROOT_ITEM,
            "connection", self,
            "id", -1,
            "name", name,
            "is-folder", TRUE,
            NULL));
    g_free (name);
    return item;
}
