/* -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*-
 *
 * Implmentation of DAAP (e.g., iTunes Music) sharing
 *
 * Copyright (C) 2005 Charles Schmidt <cschmidt2@emich.edu>
 *
 * Modifications Copyright (C) 2008 W. Michael Petullo <mike@flyn.org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 *
 */

#include "config.h"

#include <time.h>
#include <string.h>
#include <stdlib.h>

#include <glib/gi18n.h>

#include <libsoup/soup.h>
#include <libsoup/soup-address.h>
#include <libsoup/soup-message.h>
#include <libsoup/soup-uri.h>
#include <libsoup/soup-server.h>

#include <libdmapsharing/daap-record.h>
#include <libdmapsharing/daap-share.h>
#include <libdmapsharing/dmap-structure.h>
#include <libdmapsharing/dmap-container-db.h>
#include <libdmapsharing/dmap-container-record.h>

static void daap_share_set_property  (GObject *object,
					 guint prop_id,
					 const GValue *value,
					 GParamSpec *pspec);
static void daap_share_get_property  (GObject *object,
					 guint prop_id,
					 GValue *value,
				 	 GParamSpec *pspec);
static void daap_share_dispose	(GObject *object);
guint daap_share_get_desired_port (DMAPShare *share);
const char *daap_share_get_type_of_service (DMAPShare *share);
void daap_share_server_info (DMAPShare         *share,
			     SoupServer        *server,
	  		     SoupMessage       *message,
			     const char        *path,
			     GHashTable        *query,
			     SoupClientContext *context);
void daap_share_databases (DMAPShare         *share,
			     SoupServer        *server,
	  		     SoupMessage       *message,
			     const char        *path,
			     GHashTable        *query,
			     SoupClientContext *context);
void daap_share_message_add_standard_headers (SoupMessage *message);

#define DAAP_TYPE_OF_SERVICE "_daap._tcp"
#define DAAP_PORT 3689

struct DAAPSharePrivate {
	/* db things */
	DMAPDb *db;
	DMAPContainerDb *container_db;
};

#define DAAP_SHARE_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), TYPE_DAAP_SHARE, DAAPSharePrivate))

enum {
	PROP_0,
	PROP_DB,
	PROP_CONTAINER_DB,
	PROP_ENTRY_TYPE
};

/* Provide two items as user_data to callback. */
typedef struct ChunkData {
	SoupServer *server;
	GInputStream *stream;
} ChunkData;

G_DEFINE_TYPE (DAAPShare, daap_share, TYPE_DMAP_SHARE)

static void
daap_share_class_init (DAAPShareClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS (klass);
	DMAPShareClass *parent_class = DMAP_SHARE_CLASS (object_class);

	object_class->get_property = daap_share_get_property;
	object_class->set_property = daap_share_set_property;
	object_class->dispose = daap_share_dispose;

	parent_class->get_desired_port    = daap_share_get_desired_port;
	parent_class->get_type_of_service = daap_share_get_type_of_service;
	parent_class->message_add_standard_headers = daap_share_message_add_standard_headers;
	parent_class->server_info         = daap_share_server_info;
	parent_class->databases           = daap_share_databases;

	/* FIXME?: */
	g_object_class_install_property (object_class,
                                         PROP_DB,
                                         g_param_spec_pointer ("db",
                                                              "DB",
                                                              "DB object",
                                                               G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

	g_object_class_install_property (object_class,
                                         PROP_CONTAINER_DB,
                                         g_param_spec_pointer ("container-db",
                                                              "Container DB",
                                                              "Container DB object",
                                                               G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

	/* FIXME?: */
	g_object_class_install_property (object_class,
                                         PROP_ENTRY_TYPE,
                                         g_param_spec_pointer ("entry-type",
                                                             "Entry type",
                                                             "Type of entries to be shared",
                                                             G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));

	g_type_class_add_private (klass, sizeof (DAAPSharePrivate));
}

static void
daap_share_init (DAAPShare *share)
{
	share->priv = DAAP_SHARE_GET_PRIVATE (share);
	/* FIXME: do I need to manually call parent _init? */
}

static void
daap_share_set_property (GObject *object,
			    guint prop_id,
			    const GValue *value,
			    GParamSpec *pspec)
{
	DAAPShare *share = DAAP_SHARE (object);

	switch (prop_id) {
	/* FIXME: */
	case PROP_DB:
		share->priv->db = (DMAPDb *) g_value_get_pointer (value);
		break;
	case PROP_CONTAINER_DB:
		share->priv->container_db = (DMAPContainerDb *) g_value_get_pointer (value);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
daap_share_get_property (GObject *object,
			    guint prop_id,
			    GValue *value,
			    GParamSpec *pspec)
{
	DAAPShare *share = DAAP_SHARE (object);

	switch (prop_id) {
	case PROP_DB:
		g_value_set_pointer (value, share->priv->db);
		break;
	case PROP_CONTAINER_DB:
		g_value_set_pointer (value, share->priv->container_db);
		break;
	default:
		G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
		break;
	}
}

static void
daap_share_dispose (GObject *object)
{
	/* FIXME: implement in parent */
}

DAAPShare *
daap_share_new (const char *name,
		const char *password,
		DMAPDb *db,
		DMAPContainerDb *container_db)
{
	DAAPShare *share;

	share = DAAP_SHARE (g_object_new (TYPE_DAAP_SHARE,
					     "name", name,
					     "password", password,
					     "db", db,
					     "container-db", container_db,
					     NULL));

	dmap_share_server_start (DMAP_SHARE (share));
	dmap_share_publish_start (DMAP_SHARE (share));

	return share;
}

void
daap_share_message_add_standard_headers (SoupMessage *message)
{
	soup_message_headers_append (message->response_headers, "DMAP-Server", "libdmapsharing" VERSION);
}

#define DAAP_STATUS_OK 200

#define DMAP_VERSION 2.0
#define DAAP_VERSION 3.0
#define DAAP_TIMEOUT 1800

guint
daap_share_get_desired_port (DMAPShare *share)
{
	return DAAP_PORT;
}

const char *
daap_share_get_type_of_service (DMAPShare *share)
{
	return DAAP_TYPE_OF_SERVICE;
}

void
daap_share_server_info (DMAPShare *share,
		SoupServer        *server,
	  	SoupMessage       *message,
		const char        *path,
		GHashTable        *query,
		SoupClientContext *context)
{
/* MSRV	server info response
 * 	MSTT status
 * 	MPRO daap version
 * 	APRO daap version
 * 	MINM name
 * 	MSAU authentication method
 * 	MSLR login required
 * 	MSTM timeout interval
 * 	MSAL supports auto logout
 * 	MSUP supports update
 * 	MSPI supports persistent ids
 * 	MSEX supports extensions
 * 	MSBR supports browse
 * 	MSQY supports query
 * 	MSIX supports index
 * 	MSRS supports resolve
 * 	MSDC databases count
 */
	gchar *nameprop;
	GNode *msrv;

	g_warning ("path is %s\n", path);

	g_object_get ((gpointer) share, "name", &nameprop, NULL);

	msrv = dmap_structure_add (NULL, DMAP_CC_MSRV);
	dmap_structure_add (msrv, DMAP_CC_MSTT, (gint32) DAAP_STATUS_OK);
	dmap_structure_add (msrv, DMAP_CC_MPRO, (gdouble) DAAP_VERSION);
	dmap_structure_add (msrv, DMAP_CC_APRO, (gdouble) DAAP_VERSION);
	/* 2/3 is for itunes 4.8 (at least).  its determined by the
	 * Client-DAAP-Version header sent, but if we decide not to support
	 * older versions..? anyway
	 *
	 * 1.0 is 1/1
	 * 2.0 is 1/2
	 * 3.0 is 2/3
	 */
	dmap_structure_add (msrv, DMAP_CC_MINM, nameprop);
	dmap_structure_add (msrv, DMAP_CC_MSAU, dmap_share_get_auth_method (share));
	/* authentication method
	 * 0 is nothing
	 * 1 is name & password
	 * 2 is password only
	 */
	dmap_structure_add (msrv, DMAP_CC_MSLR, 0);
	dmap_structure_add (msrv, DMAP_CC_MSTM, (gint32) DAAP_TIMEOUT);
	dmap_structure_add (msrv, DMAP_CC_MSAL, (gchar) 0);
	dmap_structure_add (msrv, DMAP_CC_MSUP, (gchar) 0);
	dmap_structure_add (msrv, DMAP_CC_MSPI, (gchar) 0);
	dmap_structure_add (msrv, DMAP_CC_MSEX, (gchar) 0);
	dmap_structure_add (msrv, DMAP_CC_MSBR, (gchar) 0);
	dmap_structure_add (msrv, DMAP_CC_MSQY, (gchar) 0);
	dmap_structure_add (msrv, DMAP_CC_MSIX, (gchar) 0);
	dmap_structure_add (msrv, DMAP_CC_MSRS, (gchar) 0);
	dmap_structure_add (msrv, DMAP_CC_MSDC, (gint32) 1);

	dmap_share_message_set_from_dmap_structure (share, message, msrv);
	dmap_structure_destroy (msrv);

	g_free (nameprop);
}

typedef enum {
	ITEM_ID = 0,
	ITEM_NAME,
	ITEM_KIND,
	PERSISTENT_ID,
	CONTAINER_ITEM_ID,
	SONG_ALBUM,
	SONG_GROUPING,
	SONG_ARTIST,
	SONG_BITRATE,
	SONG_BPM,
	SONG_COMMENT,
	SONG_COMPILATION,
	SONG_COMPOSER,
	SONG_DATA_KIND,
	SONG_DATA_URL,
	SONG_DATE_ADDED,
	SONG_DATE_MODIFIED,
	SONG_DISC_COUNT,
	SONG_DISC_NUMBER,
	SONG_DISABLED,
	SONG_EQ_PRESET,
	SONG_FORMAT,
	SONG_GENRE,
	SONG_DESCRIPTION,
	SONG_RELATIVE_VOLUME,
	SONG_SAMPLE_RATE,
	SONG_SIZE,
	SONG_START_TIME,
	SONG_STOP_TIME,
	SONG_TIME,
	SONG_TRACK_COUNT,
	SONG_TRACK_NUMBER,
	SONG_USER_RATING,
	SONG_YEAR,
	SONG_HAS_VIDEO
} DAAPMetaData;

static struct DMAPMetaDataMap meta_data_map[] = {
	{"dmap.itemid",			ITEM_ID},
    	{"dmap.itemname",		ITEM_NAME},
    	{"dmap.itemkind",		ITEM_KIND},
    	{"dmap.persistentid",		PERSISTENT_ID},
	{"dmap.containeritemid",	CONTAINER_ITEM_ID},
    	{"daap.songalbum",		SONG_ALBUM},
    	{"daap.songartist",		SONG_ARTIST},
    	{"daap.songbitrate",		SONG_BITRATE},
    	{"daap.songbeatsperminute",	SONG_BPM},
    	{"daap.songcomment",		SONG_COMMENT},
    	{"daap.songcompilation",	SONG_COMPILATION},
    	{"daap.songcomposer",		SONG_COMPOSER},
    	{"daap.songdatakind",		SONG_DATA_KIND},
    	{"daap.songdataurl",		SONG_DATA_URL},
    	{"daap.songdateadded",		SONG_DATE_ADDED},
    	{"daap.songdatemodified",	SONG_DATE_MODIFIED},
    	{"daap.songdescription",	SONG_DESCRIPTION},
    	{"daap.songdisabled",		SONG_DISABLED},
    	{"daap.songdisccount",		SONG_DISC_COUNT},
    	{"daap.songdiscnumber",		SONG_DISC_NUMBER},
    	{"daap.songeqpreset",		SONG_EQ_PRESET},
    	{"daap.songformat",		SONG_FORMAT},
    	{"daap.songgenre",		SONG_GENRE},
    	{"daap.songgrouping",		SONG_GROUPING},
    	{"daap.songrelativevolume",	SONG_RELATIVE_VOLUME},
    	{"daap.songsamplerate",		SONG_SAMPLE_RATE},
    	{"daap.songsize",		SONG_SIZE},
    	{"daap.songstarttime",		SONG_START_TIME},
    	{"daap.songstoptime",		SONG_STOP_TIME},
   	{"daap.songtime",		SONG_TIME},
    	{"daap.songtrackcount",		SONG_TRACK_COUNT},
    	{"daap.songtracknumber",	SONG_TRACK_NUMBER},
    	{"daap.songuserrating",		SONG_USER_RATING},
    	{"daap.songyear",		SONG_YEAR},
	{"com.apple.itunes.has-video",	SONG_HAS_VIDEO}};

#define DAAP_ITEM_KIND_AUDIO 2
#define DAAP_SONG_DATA_KIND_NONE 0

/* FIXME: Not needed?
static void
mapped_file_message_finished (SoupMessage *message, GMappedFile *file)
{
	g_warning ("finished sending mmapped file");
	g_mapped_file_free (file);
}
*/

/* FIXME: Not needed? Will iTunes work with only chunked I/O?
static void
send_mapped_file (SoupServer *server, SoupMessage *message, DAAPRecord *record, guint64 file_size, guint64 offset)
{
        GFile *file;
        GMappedFile *mapped_file;
        char *path;
        GError *error = NULL;

        file = g_file_new_for_uri (daap_record_get_location (record));
        path = g_file_get_path (file);
        if (path == NULL) {
                g_warning ("couldn't send %s mmapped: couldn't get path", path);
                soup_message_set_status (message, SOUP_STATUS_INTERNAL_SERVER_ERROR);
                g_object_unref (file);
                return;
        }
        g_object_unref (file);
        g_warning ("sending file %s mmapped, from offset %" G_GUINT64_FORMAT, path, offset);

        mapped_file = g_mapped_file_new (path, FALSE, &error);
        if (mapped_file == NULL) {
                g_warning ("Unable to map file %s: %s", path, error->message);
                soup_message_set_status (message, SOUP_STATUS_INTERNAL_SERVER_ERROR);
        } else {
                soup_message_set_response (message, "application/x-daap-tagged",
                                           SOUP_MEMORY_TEMPORARY,
                                           g_mapped_file_get_contents (mapped_file) + offset,
                                           file_size);

                g_signal_connect (message,
                                  "finished",
                                  G_CALLBACK (mapped_file_message_finished),
                                  mapped_file);
        }
        g_free (path);
}
*/

static void
write_next_chunk (SoupMessage *message, ChunkData *cd)
{
	gssize read_size;
	GError *error = NULL;
	gchar *chunk = g_malloc (DMAP_SHARE_CHUNK_SIZE);

	read_size = g_input_stream_read (cd->stream,
					 chunk,
					 DMAP_SHARE_CHUNK_SIZE,
					 NULL,
					 &error);
	if (read_size > 0) {
		soup_message_body_append (message->response_body,
					  SOUP_MEMORY_TAKE,
					  chunk,
					  read_size);
		g_warning ("wrote %" G_GSSIZE_FORMAT " bytes", read_size);
	} else {
		if (error != NULL) {
			g_warning ("error reading from input stream: %s",
				   error->message);
			g_error_free (error);
		}
		g_free (chunk);
		g_warning ("wrote 0 bytes, sending message complete");
		soup_message_body_complete (message->response_body);
	}
	soup_server_unpause_message (cd->server, message);
}

static void
chunked_message_finished (SoupMessage *message, ChunkData *cd)
{
	g_warning ("finished sending chunked file");
	g_input_stream_close (cd->stream, NULL, NULL);
	g_free (cd);
}

static void
send_chunked_file (SoupServer *server, SoupMessage *message, DAAPRecord *record, guint64 file_size, guint64 offset)
{
	const char *location;
	GError *error = NULL;
	ChunkData *cd = g_new (ChunkData, 1);

	location = daap_record_get_location (record);

	g_warning ("sending %s chunked from offset %" G_GUINT64_FORMAT, location, offset);

	cd->server = server;
	cd->stream = G_INPUT_STREAM (daap_record_read (record, &error));

	if (error != NULL) {
		g_warning ("couldn't open %s: %s", location, error->message);
		g_error_free (error);
		soup_message_set_status (message, SOUP_STATUS_INTERNAL_SERVER_ERROR);
		return;
	}

	if (offset != 0) {
		if (g_seekable_seek (G_SEEKABLE (cd->stream), offset, G_SEEK_SET, NULL, &error) == FALSE) {
			g_warning ("error seeking: %s", error->message);
			g_input_stream_close (cd->stream, NULL, NULL);
			soup_message_set_status (message, SOUP_STATUS_INTERNAL_SERVER_ERROR);
			return;
	 	}
		file_size -= offset;
	}

	/* Free memory after each chunk sent out over network. */
	soup_message_body_set_accumulate (message->response_body, FALSE);

	if (daap_record_has_video (DAAP_RECORD (record))) {
		/* NOTE: iTunes seems to require this or it stops reading 
		 * video data after about 2.5MB. Perhaps this is so iTunes
		 * knows how much data to buffer.
		 */
		soup_message_headers_set_encoding (message->response_headers,
						  SOUP_ENCODING_CONTENT_LENGTH);
		soup_message_headers_set_content_length (
			message->response_headers, file_size);
	} else if (soup_message_get_http_version (message) == SOUP_HTTP_1_0) {
#ifdef HAVE_ENCODING_EOF
		soup_message_headers_set_encoding (message->response_headers,
						   SOUP_ENCODING_EOF);
#else
		g_warning ("Received HTTP 1.0 request, but not built with HTTP 1.0 support");
		soup_message_headers_set_encoding (message->response_headers,
						   SOUP_ENCODING_CHUNKED);
#endif
	} else {
		soup_message_headers_set_encoding (message->response_headers,
						   SOUP_ENCODING_CHUNKED);
	}

	soup_message_headers_append (message->response_headers, "Connection", "Close");
	soup_message_headers_append (message->response_headers, "Content-Type", "application/x-daap-tagged");

	g_signal_connect (message, "wrote_headers", G_CALLBACK (write_next_chunk), cd);
	g_signal_connect (message, "wrote_chunk", G_CALLBACK (write_next_chunk), cd);
	g_signal_connect (message, "finished", G_CALLBACK (chunked_message_finished), cd);
}

static void
add_entry_to_mlcl (gint *id, DMAPRecord *record, gpointer _mb)
{
	struct MLCL_Bits *mb = (struct MLCL_Bits *) _mb;
	GNode *mlit;
	mlit = dmap_structure_add (mb->mlcl, DMAP_CC_MLIT);

	if (dmap_share_client_requested (mb->bits, ITEM_KIND))
		dmap_structure_add (mlit, DMAP_CC_MIKD, (gchar) DAAP_ITEM_KIND_AUDIO);
	if (dmap_share_client_requested (mb->bits, ITEM_ID))
		dmap_structure_add (mlit, DMAP_CC_MIID, (gint32) *id);
	if (dmap_share_client_requested (mb->bits, ITEM_NAME))
		dmap_structure_add (mlit, DMAP_CC_MINM, daap_record_get_title (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, PERSISTENT_ID))
		dmap_structure_add (mlit, DMAP_CC_MPER, (gint64) *id);
	if (dmap_share_client_requested (mb->bits, CONTAINER_ITEM_ID))
		dmap_structure_add (mlit, DMAP_CC_MCTI, id);
	if (dmap_share_client_requested (mb->bits, SONG_DATA_KIND))
		dmap_structure_add (mlit, DMAP_CC_ASDK, (gchar) DAAP_SONG_DATA_KIND_NONE);
	if (dmap_share_client_requested (mb->bits, SONG_DATA_URL))
		dmap_structure_add (mlit, DMAP_CC_ASUL, "");
	if (dmap_share_client_requested (mb->bits, SONG_ALBUM))
		dmap_structure_add (mlit, DMAP_CC_ASAL, daap_record_get_album (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, SONG_GROUPING))
		dmap_structure_add (mlit, DMAP_CC_AGRP, "");
	if (dmap_share_client_requested (mb->bits, SONG_ARTIST))
		dmap_structure_add (mlit, DMAP_CC_ASAR, daap_record_get_artist (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, SONG_BITRATE)) {
		gulong bitrate = daap_record_get_bitrate (DAAP_RECORD (record));
		if (bitrate != 0)
			dmap_structure_add (mlit, DMAP_CC_ASBR, (gint32) bitrate);
	}
	if (dmap_share_client_requested (mb->bits, SONG_BPM))
		dmap_structure_add (mlit, DMAP_CC_ASBT, (gint32) 0);
	if (dmap_share_client_requested (mb->bits, SONG_COMMENT))
		dmap_structure_add (mlit, DMAP_CC_ASCM, "");
	if (dmap_share_client_requested (mb->bits, SONG_COMPILATION))
		dmap_structure_add (mlit, DMAP_CC_ASCO, (gchar) FALSE);
	if (dmap_share_client_requested (mb->bits, SONG_COMPOSER))
		dmap_structure_add (mlit, DMAP_CC_ASCP, "");
	if (dmap_share_client_requested (mb->bits, SONG_DATE_ADDED))
		dmap_structure_add (mlit, DMAP_CC_ASDA, daap_record_get_firstseen (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, SONG_DATE_MODIFIED))
		dmap_structure_add (mlit, DMAP_CC_ASDM, daap_record_get_mtime (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, SONG_DISC_COUNT))
		dmap_structure_add (mlit, DMAP_CC_ASDC, (gint32) 0);
	if (dmap_share_client_requested (mb->bits, SONG_DISC_NUMBER))
		dmap_structure_add (mlit, DMAP_CC_ASDN, daap_record_get_disc (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, SONG_DISABLED))
		dmap_structure_add (mlit, DMAP_CC_ASDB, (gchar) FALSE);
	if (dmap_share_client_requested (mb->bits, SONG_EQ_PRESET))
		dmap_structure_add (mlit, DMAP_CC_ASEQ, "");
	if (dmap_share_client_requested (mb->bits, SONG_FORMAT))
		dmap_structure_add (mlit, DMAP_CC_ASFM, daap_record_get_format (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, SONG_GENRE))
		dmap_structure_add (mlit, DMAP_CC_ASGN, daap_record_get_genre (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, SONG_DESCRIPTION))
		dmap_structure_add (mlit, DMAP_CC_ASDT, ""); /* FIXME: e.g., wav audio file */
	if (dmap_share_client_requested (mb->bits, SONG_RELATIVE_VOLUME))
		dmap_structure_add (mlit, DMAP_CC_ASRV, 0);
	if (dmap_share_client_requested (mb->bits, SONG_SAMPLE_RATE))
		dmap_structure_add (mlit, DMAP_CC_ASSR, 0);
	if (dmap_share_client_requested (mb->bits, SONG_SIZE))
		dmap_structure_add (mlit, DMAP_CC_ASSZ, (gint32) daap_record_get_filesize (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, SONG_START_TIME))
		dmap_structure_add (mlit, DMAP_CC_ASST, 0);
	if (dmap_share_client_requested (mb->bits, SONG_STOP_TIME))
		dmap_structure_add (mlit, DMAP_CC_ASSP, 0);
	if (dmap_share_client_requested (mb->bits, SONG_TIME))
		dmap_structure_add (mlit, DMAP_CC_ASTM, (1000 * daap_record_get_duration (DAAP_RECORD (record))));
	if (dmap_share_client_requested (mb->bits, SONG_TRACK_COUNT))
		dmap_structure_add (mlit, DMAP_CC_ASTC, 0);
	if (dmap_share_client_requested (mb->bits, SONG_TRACK_NUMBER))
		dmap_structure_add (mlit, DMAP_CC_ASTN, (gint32) daap_record_get_track (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, SONG_USER_RATING))
		dmap_structure_add (mlit, DMAP_CC_ASUR, 0); /* FIXME */
	if (dmap_share_client_requested (mb->bits, SONG_YEAR))
		dmap_structure_add (mlit, DMAP_CC_ASYR, daap_record_get_year (DAAP_RECORD (record)));
	if (dmap_share_client_requested (mb->bits, SONG_HAS_VIDEO))
		dmap_structure_add (mlit, DMAP_CC_AEHV, daap_record_has_video (DAAP_RECORD (record)));

	return;
}

void
daap_share_databases (DMAPShare *share,
	      SoupServer        *server,
	      SoupMessage       *message,
	      const char        *path,
	      GHashTable        *query,
	      SoupClientContext *context)

{
	const char *rest_of_path;
	/*guint revision_number;*/

	g_warning ("path is %s\n", path);
	g_warning ("type is %s\n", (char *) g_hash_table_lookup (query, "type"));
	g_warning ("meta is %s\n", (char *) g_hash_table_lookup (query, "meta"));
	g_warning ("session-id is %s\n", (char *) g_hash_table_lookup (query, "session-id"));
	g_warning ("revision-number is %s\n", (char *) g_hash_table_lookup (query, "revision-number"));

	if (! dmap_share_session_id_validate (share, context, message, query, NULL)) {
		soup_message_set_status (message, SOUP_STATUS_FORBIDDEN);
		return;
	}

	rest_of_path = strchr (path + 1, '/');

	if (rest_of_path == NULL) {
	/* AVDB server databases
	 * 	MSTT status
	 * 	MUTY update type
	 * 	MTCO specified total count
	 * 	MRCO returned count
	 * 	MLCL listing
	 * 		MLIT listing item
	 * 			MIID item id
	 * 			MPER persistent id
	 * 			MINM item name
	 * 			MIMC item count
	 * 			MCTC container count
	 */
		gchar *nameprop;
		GNode *avdb;
		GNode *mlcl;
		GNode *mlit;

		g_object_get ((gpointer) share, "name", &nameprop, NULL);

		avdb = dmap_structure_add (NULL, DMAP_CC_AVDB);
		dmap_structure_add (avdb, DMAP_CC_MSTT, (gint32) DAAP_STATUS_OK);
		dmap_structure_add (avdb, DMAP_CC_MUTY, 0);
		dmap_structure_add (avdb, DMAP_CC_MTCO, (gint32) 1);
		dmap_structure_add (avdb, DMAP_CC_MRCO, (gint32) 1);
		mlcl = dmap_structure_add (avdb, DMAP_CC_MLCL);
		mlit = dmap_structure_add (mlcl, DMAP_CC_MLIT);
		dmap_structure_add (mlit, DMAP_CC_MIID, (gint32) 1);
		dmap_structure_add (mlit, DMAP_CC_MPER, (gint64) 1);
		dmap_structure_add (mlit, DMAP_CC_MINM, nameprop);
		dmap_structure_add (mlit, DMAP_CC_MIMC, dmap_db_count (DAAP_SHARE (share)->priv->db));
		dmap_structure_add (mlit, DMAP_CC_MCTC, (gint32) 1);

		dmap_share_message_set_from_dmap_structure (share, message, avdb);
		dmap_structure_destroy (avdb);

		g_free (nameprop);
	} else if (g_ascii_strcasecmp ("/1/items", rest_of_path) == 0) {
	/* ADBS database songs
	 * 	MSTT status
	 * 	MUTY update type
	 * 	MTCO specified total count
	 * 	MRCO returned count
	 * 	MLCL listing
	 * 		MLIT
	 * 			attrs
	 * 		MLIT
	 * 		...
	 */
		GNode *adbs;
		gint32 num_songs = dmap_db_count (DAAP_SHARE (share)->priv->db);
		struct MLCL_Bits mb = {NULL,0};

		mb.bits = dmap_share_parse_meta (query, meta_data_map, G_N_ELEMENTS (meta_data_map));

		adbs = dmap_structure_add (NULL, DMAP_CC_ADBS);
		dmap_structure_add (adbs, DMAP_CC_MSTT, (gint32) DAAP_STATUS_OK);
		dmap_structure_add (adbs, DMAP_CC_MUTY, 0);
		dmap_structure_add (adbs, DMAP_CC_MTCO, (gint32) num_songs);
		dmap_structure_add (adbs, DMAP_CC_MRCO, (gint32) num_songs);
		mb.mlcl = dmap_structure_add (adbs, DMAP_CC_MLCL);

		dmap_db_foreach (DAAP_SHARE (share)->priv->db, add_entry_to_mlcl, (gpointer) &mb);

		dmap_share_message_set_from_dmap_structure (share, message, adbs);
		dmap_structure_destroy (adbs);
		adbs = NULL;
	} else if (g_ascii_strcasecmp ("/1/containers", rest_of_path) == 0) {
	/* APLY database playlists
	 * 	MSTT status
	 * 	MUTY update type
	 * 	MTCO specified total count
	 * 	MRCO returned count
	 * 	MLCL listing
	 * 		MLIT listing item
	 * 			MIID item id
	 * 			MPER persistent item id
	 * 			MINM item name
	 * 			MIMC item count
	 * 			ABPL baseplaylist (only for base)
	 * 		MLIT
	 * 		...
	 */
		gchar *nameprop;
		GNode *aply;
		GNode *mlcl;
		GNode *mlit;

		g_object_get ((gpointer) share, "name", &nameprop, NULL);

		aply = dmap_structure_add (NULL, DMAP_CC_APLY);
		dmap_structure_add (aply, DMAP_CC_MSTT, (gint32) DAAP_STATUS_OK);
		dmap_structure_add (aply, DMAP_CC_MUTY, 0);
		dmap_structure_add (aply, DMAP_CC_MTCO, (gint32) dmap_container_db_count (DAAP_SHARE (share)->priv->container_db) + 1);
		dmap_structure_add (aply, DMAP_CC_MRCO, (gint32) dmap_container_db_count (DAAP_SHARE (share)->priv->container_db) + 1);
		mlcl = dmap_structure_add (aply, DMAP_CC_MLCL);

		/* Base playlist: */
		mlit = dmap_structure_add (mlcl, DMAP_CC_MLIT);
		dmap_structure_add (mlit, DMAP_CC_MIID, (gint32) 1);
		dmap_structure_add (mlit, DMAP_CC_MPER, (gint64) 1);
		dmap_structure_add (mlit, DMAP_CC_MINM, nameprop);
		dmap_structure_add (mlit, DMAP_CC_MIMC, dmap_db_count (DAAP_SHARE (share)->priv->db));
		dmap_structure_add (mlit, DMAP_CC_ABPL, (gchar) 1);

		dmap_container_db_foreach (DAAP_SHARE (share)->priv->container_db, dmap_share_add_playlist_to_mlcl, (gpointer) mlcl);

		dmap_share_message_set_from_dmap_structure (share, message, aply);
		dmap_structure_destroy (aply);

		g_free (nameprop);
	} else if (g_ascii_strncasecmp ("/1/containers/", rest_of_path, 14) == 0) {
	/* APSO playlist songs
	 * 	MSTT status
	 * 	MUTY update type
	 * 	MTCO specified total count
	 * 	MRCO returned count
	 * 	MLCL listing
	 * 		MLIT listing item
	 * 			MIKD item kind
	 * 			MIID item id
	 * 			MCTI container item id
	 * 		MLIT
	 * 		...
	 */
		GNode *apso;
		struct MLCL_Bits mb = {NULL,0};
		gint pl_id = atoi (rest_of_path + 14);

		mb.bits = dmap_share_parse_meta (query, meta_data_map, G_N_ELEMENTS (meta_data_map));

		apso = dmap_structure_add (NULL, DMAP_CC_APSO);
		dmap_structure_add (apso, DMAP_CC_MSTT, (gint32) DAAP_STATUS_OK);
		dmap_structure_add (apso, DMAP_CC_MUTY, 0);

		if (pl_id == 1) {
			gint32 num_songs = dmap_db_count (DAAP_SHARE (share)->priv->db);
			dmap_structure_add (apso, DMAP_CC_MTCO, (gint32) num_songs);
			dmap_structure_add (apso, DMAP_CC_MRCO, (gint32) num_songs);
			mb.mlcl = dmap_structure_add (apso, DMAP_CC_MLCL);

			dmap_db_foreach (DAAP_SHARE (share)->priv->db, add_entry_to_mlcl, (gpointer) &mb);
		} else {
			DMAPContainerRecord *record;
			DMAPDb *entries;
			guint num_songs;
			
			record = dmap_container_db_lookup_by_id (DAAP_SHARE (share)->priv->container_db, pl_id);
			entries = dmap_container_record_get_entries (record,
								     DAAP_SHARE (share)->priv->db);
			num_songs = dmap_db_count (entries);
			
			dmap_structure_add (apso, DMAP_CC_MTCO, (gint32) num_songs);
			dmap_structure_add (apso, DMAP_CC_MRCO, (gint32) num_songs);
			mb.mlcl = dmap_structure_add (apso, DMAP_CC_MLCL);

			dmap_db_foreach (entries, add_entry_to_mlcl, (gpointer) &mb);
		}

		dmap_share_message_set_from_dmap_structure (share, message, apso);
		dmap_structure_destroy (apso);
	} else if (g_ascii_strncasecmp ("/1/items/", rest_of_path, 9) == 0) {
	/* just the file :) */
		const gchar *id_str;
		gint id;
		const gchar *location;
		const gchar *range_header;
		guint64 file_size;
		guint64 offset = 0;
		DAAPRecord *record;

		id_str = rest_of_path + 9;
		id = atoi (id_str);

		record = DAAP_RECORD (dmap_db_lookup_by_id (DAAP_SHARE (share)->priv->db, id));
		location = daap_record_get_location (record);
		file_size = daap_record_get_filesize (record);

		daap_share_message_add_standard_headers (message);
		soup_message_headers_append (message->response_headers, "Accept-Ranges", "bytes");

		range_header = soup_message_headers_get (message->request_headers, "Range");
		if (range_header) {
			const gchar *s;
			gchar *content_range;

			s = range_header + 6; /* bytes= */
			offset = atoll (s);

			content_range = g_strdup_printf ("bytes %" G_GUINT64_FORMAT "-%" G_GUINT64_FORMAT "/%" G_GUINT64_FORMAT, offset, file_size, file_size);
			soup_message_headers_append (message->response_headers, "Content-Range", content_range);
			g_free (content_range);

			soup_message_set_status (message, SOUP_STATUS_PARTIAL_CONTENT);
			file_size -= offset;
		} else {
			soup_message_set_status (message, SOUP_STATUS_OK);
		}

		/* don't use chunked transfers if we can send the file mmapped,
		 * as itunes clients can't seek properly when we do. Actually, I'm
		 * not sure this is true. iTunes 8 seeks fine. We need to use chunked
		 * transfers to support transcoding.
		 if (file_size < 1024 * 1024 * 10 // 10MB file too big to mmap
		     && dmap_share_uri_is_local (location)) {
		  	send_mapped_file (server, message, record, file_size, offset);
		 } else {
		 	send_chunked_file (server, message, record, file_size, offset);
		 }
		 */
		 
		send_chunked_file (server, message, record, file_size, offset);
	} else {
		g_warning ("unhandled: %s\n", path);
	}
}
