/*
 * Copyright (C) 2004,2005 Charles Schmidt <cschmidt2@emich.edu>
 * Copyright (C) 2006 INDT
 *  Andre Moreira Magalhaes <andre.magalhaes@indt.org.br>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "dmap-connection.h"
#include "dmap-marshal.h"
#include "dmap-structure.h"

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_LIBZ
#include <zlib.h>
#endif

#include <libsoup/soup.h>

typedef struct _DMAPResponseHandlerData DMAPResponseHandlerData;

struct _DMAPResponseHandlerData {
    DMAPConnection *self;
    DMAPResponseHandler response_handler;
    gpointer response_user_data;
};

struct _DMAPConnectionPrivate {
    gchar *name;
    gchar *username;
    gchar *password;
    gchar *host;
    guint port;
    gboolean need_revision_number;

    SoupSession *session;
    SoupURI *base_uri;

    gdouble protocol_version;
    guint32 session_id;
    gint database_id;
    gint revision_number;
    DMAPItem *root_item;

    gboolean result;

    enum {
        DMAP_GET_INFO = 0,
        DMAP_GET_PASSWORD,
        DMAP_LOGIN,
        DMAP_GET_REVISION_NUMBER,
        DMAP_GET_DB_INFO,
        DMAP_CONNECTED,
        DMAP_LOGOUT,
        DMAP_DONE,
    } state;
};

enum {
    PROP_0,
    PROP_NAME,
    PROP_USERNAME,
    PROP_PASSWORD,
    PROP_HOST,
    PROP_PORT,
    PROP_NEED_REVISION_NUMBER,
    PROP_ROOT_ITEM,
};

enum {
    CONNECTED,
    DISCONNECTED,
    ERROR,
    LAST_SIGNAL
};

static void dmap_connection_dispose (GObject *obj);
static void dmap_connection_finalize (GObject *obj);
static void dmap_connection_set_property (GObject *object,
                                          guint prop_id,
                                          const GValue *value,
                                          GParamSpec *pspec);
static void dmap_connection_get_property (GObject *object,
                                          guint prop_id,
                                          GValue *value,
                                          GParamSpec *pspec);

static void dmap_connection_do_something (DMAPConnection *self);
static void dmap_connection_state_done (DMAPConnection *self,
                                        gboolean result,
                                        DMAPConnectionError error);
static DMAPConnectionError dmap_connection_map_error (guint status);

G_DEFINE_TYPE (DMAPConnection, dmap_connection, G_TYPE_OBJECT)

#define DMAP_CONNECTION_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE ((o), TYPE_DMAP_CONNECTION, DMAPConnectionPrivate))

static guint dmap_connection_signals [LAST_SIGNAL] = { 0, };

static void
dmap_connection_class_init (DMAPConnectionClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->dispose = dmap_connection_dispose;
    object_class->finalize = dmap_connection_finalize;
    object_class->set_property = dmap_connection_set_property;
    object_class->get_property = dmap_connection_get_property;

    g_type_class_add_private (klass, sizeof (DMAPConnectionPrivate));

    g_object_class_install_property (object_class,
            PROP_NAME,
            g_param_spec_string ("name",
                "self name",
                "self name",
                NULL,
                G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
    g_object_class_install_property (object_class,
            PROP_USERNAME,
            g_param_spec_string ("username",
                "self username",
                "self username",
                NULL,
                G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
    g_object_class_install_property (object_class,
            PROP_PASSWORD,
            g_param_spec_string ("password",
                "connnection password",
                "self password",
                NULL,
                G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
    g_object_class_install_property (object_class,
            PROP_HOST,
            g_param_spec_string ("host",
                "host",
                "host",
                NULL,
                G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
    g_object_class_install_property (object_class,
            PROP_PORT,
            g_param_spec_uint ("port",
                "port",
                "port",
                0, G_MAXINT, 0,
                G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
    g_object_class_install_property (object_class,
            PROP_NEED_REVISION_NUMBER,
            g_param_spec_boolean ("need-revision-number",
                "Need revision number",
                "Need revision number",
                FALSE,
                G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY));
    g_object_class_install_property (object_class,
            PROP_ROOT_ITEM,
            g_param_spec_object ("root-item",
                "self root item",
                "self root item",
                TYPE_DMAP_ITEM,
                G_PARAM_READABLE));

    dmap_connection_signals [CONNECTED] = g_signal_new ("connected",
            G_TYPE_FROM_CLASS (object_class),
            G_SIGNAL_RUN_LAST,
            G_STRUCT_OFFSET (DMAPConnectionClass, connected),
            NULL,
            NULL,
            g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);
    dmap_connection_signals [DISCONNECTED] = g_signal_new ("disconnected",
            G_TYPE_FROM_CLASS (object_class),
            G_SIGNAL_RUN_LAST,
            G_STRUCT_OFFSET (DMAPConnectionClass, disconnected),
            NULL,
            NULL,
            g_cclosure_marshal_VOID__VOID,
            G_TYPE_NONE,
            0);
    dmap_connection_signals [ERROR] = g_signal_new ("error",
            G_TYPE_FROM_CLASS (object_class),
            G_SIGNAL_RUN_LAST,
            G_STRUCT_OFFSET (DMAPConnectionClass, error),
            NULL,
            NULL,
            g_cclosure_marshal_VOID__INT,
            G_TYPE_NONE,
            1,
            G_TYPE_INT);
}

static void
dmap_connection_init (DMAPConnection *self)
{
    self->priv = DMAP_CONNECTION_GET_PRIVATE (self);
    memset (self->priv, 0, sizeof (DMAPConnectionPrivate));
    self->priv->username = g_strdup_printf ("libdmapsharing_%s", VERSION);
}

static void
dmap_connection_dispose (GObject *object)
{
    DMAPConnectionPrivate *priv = DMAP_CONNECTION (object)->priv;

    dmap_connection_disconnect (DMAP_CONNECTION (object));

    if (priv->root_item)
        g_object_unref (priv->root_item);

    g_free (priv->name);
    g_free (priv->username);
    g_free (priv->password);
    g_free (priv->host);

    if (priv->session) {
        soup_session_abort (priv->session);
        g_object_unref (priv->session);
        priv->session = NULL;
    }

    if (priv->base_uri) {
        soup_uri_free (priv->base_uri);
        priv->base_uri = NULL;
    }

    G_OBJECT_CLASS (dmap_connection_parent_class)->dispose (object);
}

static void
dmap_connection_finalize (GObject *object)
{
    g_signal_handlers_destroy (object);
    G_OBJECT_CLASS (dmap_connection_parent_class)->finalize (object);
}

static void
dmap_connection_set_property (GObject *object,
                              guint prop_id,
                              const GValue *value,
                              GParamSpec *pspec)
{
    DMAPConnectionPrivate *priv = DMAP_CONNECTION (object)->priv;

    switch (prop_id) {
    case PROP_NAME:
        priv->name = g_value_dup_string (value);
        break;
    case PROP_USERNAME:
        priv->username = g_value_dup_string (value);
        break;
    case PROP_PASSWORD:
        priv->password = g_value_dup_string (value);
        break;
    case PROP_HOST:
        priv->host = g_value_dup_string (value);
        break;
    case PROP_PORT:
        priv->port = g_value_get_uint (value);
        break;
    case PROP_NEED_REVISION_NUMBER:
        priv->need_revision_number = g_value_get_boolean (value);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static void
dmap_connection_get_property (GObject *object,
                              guint prop_id,
                              GValue *value,
                              GParamSpec *pspec)
{
    DMAPConnectionPrivate *priv = DMAP_CONNECTION (object)->priv;

    switch (prop_id) {
    case PROP_NAME:
        g_value_set_string (value, priv->name);
        break;
    case PROP_USERNAME:
        g_value_set_string (value, priv->username);
        break;
    case PROP_PASSWORD:
        g_value_set_string (value, priv->password);
        break;
    case PROP_HOST:
        g_value_set_string (value, priv->host);
        break;
    case PROP_PORT:
        g_value_set_uint (value, priv->port);
        break;
    case PROP_NEED_REVISION_NUMBER:
        g_value_set_boolean (value, priv->need_revision_number);
        break;
    case PROP_ROOT_ITEM:
        g_value_set_object (value, priv->root_item);
        break;
    default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
        break;
    }
}

static SoupMessage *
build_message (DMAPConnection *self,
               const gchar *path,
               gboolean need_hash,
               gdouble version,
               gint req_id,
               gboolean send_close)
{
    DMAPConnectionPrivate *priv = self->priv;
    DMAPConnectionClass *klass = DMAP_CONNECTION_GET_CLASS (self);
    SoupMessage *message = NULL;

    message = klass->build_message (self, 
            priv->base_uri, path, need_hash, version, req_id);
    if (!message)
        return NULL;

#ifdef HAVE_LIBZ
    soup_message_headers_append (message->request_headers, "Accept-Encoding", "gzip");
#endif

    if (priv->password) {
        gchar *h;
        gchar *user_pass;
        gchar *token;

        user_pass = g_strdup_printf ("%s:%s", priv->username, priv->password);
        token = g_base64_encode ((guchar *) user_pass, strlen (user_pass));
        h = g_strdup_printf ("Basic %s", token);

        g_free (token);
        g_free (user_pass);

        soup_message_headers_append (message->request_headers, "Authorization", h);
        g_free (h);
    }

    if (send_close) {
        soup_message_headers_append (message->request_headers, "self", "close");
    }

    return message;
}

#ifdef HAVE_LIBZ
static void *
g_zalloc_wrapper (voidpf opaque, uInt items, uInt size)
{
    if ((items != 0) && (size >= G_MAXUINT/items)) {
        return Z_NULL;
    }
    if ((size != 0) && (items >= G_MAXUINT/size)) {
        return Z_NULL;
    }
    return g_malloc0 (items * size);
}

static void
g_zfree_wrapper (voidpf opaque, voidpf address)
{
    g_free (address);
}
#endif

static void
http_response_handler (SoupSession *session, SoupMessage *message,
                       DMAPResponseHandlerData *data)
{
    DMAPConnectionPrivate *priv;
    GNode *structure;
    guint status;
    const gchar *response;
    gchar *new_response = NULL;
    gint response_length;
    const gchar *encoding_header;
    gchar *message_path;

    priv = data->self->priv;
    structure = NULL;
    status = message->status_code;
    response = message->response_body->data;
    response_length = message->response_body->length;
    encoding_header = NULL;

    message_path = soup_uri_to_string (soup_message_get_uri (message), FALSE);

    g_debug ("Received response from %s: %d, %s",
          message_path,
          message->status_code,
          message->reason_phrase);

    if (status == SOUP_STATUS_CANCELLED) {
        return;
    }

    if (response_length >= G_MAXUINT/4 - 1) {
        /* If response_length is too big,
         * the g_malloc (unc_size + 1) below would overflow
         */
        status = SOUP_STATUS_MALFORMED;
    }

    if (message->response_headers) {
        encoding_header = soup_message_headers_get (message->response_headers, "Content-Encoding");
    }

    if (SOUP_STATUS_IS_SUCCESSFUL (status) && encoding_header && strcmp (encoding_header, "gzip") == 0) {
#ifdef HAVE_LIBZ
        z_stream stream;
        unsigned int factor = 4;
        unsigned int unc_size = response_length * factor;

        stream.next_in = (unsigned char *)response;
        stream.avail_in = response_length;
        stream.total_in = 0;

        new_response = g_malloc (unc_size + 1);
        stream.next_out = (unsigned char *)new_response;
        stream.avail_out = unc_size;
        stream.total_out = 0;
        stream.zalloc = g_zalloc_wrapper;
        stream.zfree = g_zfree_wrapper;
        stream.opaque = NULL;

        if (inflateInit2 (&stream, 32 /* auto-detect */ + 15 /* max */ ) != Z_OK) {
            inflateEnd (&stream);
            g_free (new_response);
            g_debug ("Unable to decompress response from %s",
                  message_path);
            status = SOUP_STATUS_MALFORMED;
        } else {
            do {
                gint z_res = inflate (&stream, Z_FINISH);
                if (z_res == Z_STREAM_END)
                    break;
                if ((z_res != Z_OK && z_res != Z_BUF_ERROR) || stream.avail_out != 0 || unc_size > 40*1000*1000) {
                    inflateEnd (&stream);
                    g_free (new_response);
                    new_response = NULL;
                    break;
                }

                factor *= 4;
                unc_size = (response_length * factor);
                /* unc_size can't grow bigger than 40KB, so
                 * unc_size can't overflow, and this realloc
                 * call is safe
                 */
                new_response = g_realloc (new_response, unc_size + 1);
                stream.next_out = (unsigned char *)(new_response + stream.total_out);
                stream.avail_out = unc_size - stream.total_out;
            } while (1);
        }

        if (new_response) {
            response = new_response;
            response_length = stream.total_out;
        }
#else
        g_debug ("Received compressed response from %s but can't handle it",
              message_path);
        status = SOUP_STATUS_MALFORMED;
#endif
    }

    if (SOUP_STATUS_IS_SUCCESSFUL (status)) {
        DMAPStructureItem *item;

        structure = dmap_structure_parse (response, response_length);
        if (structure == NULL) {
            g_debug ("No dmap structure returned from %s",
                  message_path);

            status = SOUP_STATUS_MALFORMED;
        } else {
            gint dmap_status = 0;
            item = dmap_structure_find_item (structure, DMAP_CC_MSTT);
            if (item)
                dmap_status = g_value_get_int (&(item->content));

            if (dmap_status != 200) {
                g_debug ("Error, dmap.status is not 200 in response from %s",
                      message_path);

                status = SOUP_STATUS_MALFORMED;
            }
        }
    } else {
        g_debug ("Error getting %s: %d, %s\n",
              message_path,
              message->status_code,
              message->reason_phrase);
    }

    if (data->response_handler) {
        DMAPResponseHandler h = data->response_handler;
        (*h) (data->self, status, structure, data->response_user_data);
    }

    g_free (data);

    if (structure) {
        dmap_structure_destroy (structure);
    }

    if (response != message->response_body->data) {
        g_free (new_response);
    }

    g_free (message_path);
}

static gboolean
http_get (DMAPConnection *self,
          const gchar *path,
          gboolean need_hash,
          gdouble version,
          gint req_id,
          gboolean send_close,
          DMAPResponseHandler handler,
          gpointer user_data)
{
    DMAPConnectionPrivate *priv = self->priv;
    SoupMessage *message;

    message = build_message (self, path, need_hash, version, req_id, send_close);
    if (message == NULL) {
        g_debug ("Error building message for http://%s:%d/%s",
              priv->base_uri->host,
              priv->base_uri->port,
              path);
        return FALSE;
    }

    DMAPResponseHandlerData *data = g_new (DMAPResponseHandlerData, 1);
    data->self = self;
    data->response_handler = handler;
    data->response_user_data = user_data;

    soup_session_queue_message (priv->session, message,
                    (SoupSessionCallback) http_response_handler,
                    data);

    g_debug ("Queued message for http://%s:%d/%s",
          priv->base_uri->host,
          priv->base_uri->port,
          path);
    return TRUE;
}

static void
handle_server_info (DMAPConnection *self,
                    guint status,
                    GNode *structure,
                    gpointer user_data)
{
    DMAPConnectionPrivate *priv = self->priv;
    DMAPStructureItem *item = NULL;

    if (!SOUP_STATUS_IS_SUCCESSFUL (status) || structure == NULL) {
        dmap_connection_state_done (self, FALSE, dmap_connection_map_error (status));
        return;
    }

    /* get the dmap version number */
    item = dmap_structure_find_item (structure, DMAP_CC_APRO);
    if (item == NULL) {
        g_debug ("apro version not found. trying ppro");
        item = dmap_structure_find_item (structure, DMAP_CC_PPRO);
        if (item == NULL) {
            g_debug ("ppro version not found. aborting");
            dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
            return;
        }
    }

    priv->protocol_version = g_value_get_double (&(item->content));
    dmap_connection_state_done (self, TRUE, DMAP_CONNECTION_ERROR_NONE);
}

static void
handle_login (DMAPConnection *self,
              guint status,
              GNode *structure,
              gpointer user_data)
{
    DMAPConnectionPrivate *priv = self->priv;
    DMAPStructureItem *item = NULL;

    if (status == SOUP_STATUS_UNAUTHORIZED || status == SOUP_STATUS_FORBIDDEN) {
        g_debug ("Incorrect password");
        priv->state = DMAP_GET_PASSWORD;
        dmap_connection_do_something (self);
    }

    if (structure == NULL || SOUP_STATUS_IS_SUCCESSFUL (status) == FALSE) {
        dmap_connection_state_done (self, FALSE, dmap_connection_map_error (status));
        return;
    }

    item = dmap_structure_find_item (structure, DMAP_CC_MLID);
    if (item == NULL) {
        g_debug ("Could not find dmap.sessionid item in /login");
        dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
        return;
    }

    priv->session_id = (guint32) g_value_get_int (&(item->content));
    dmap_connection_state_done (self, TRUE, DMAP_CONNECTION_ERROR_NONE);
}

static void
handle_update (DMAPConnection *self,
               guint status,
               GNode *structure,
               gpointer user_data)
{
    DMAPConnectionPrivate *priv = self->priv;
    DMAPStructureItem *item;

    if (structure == NULL || SOUP_STATUS_IS_SUCCESSFUL (status) == FALSE) {
        dmap_connection_state_done (self, FALSE, dmap_connection_map_error (status));
        return;
    }

    /* get a revision number */
    item = dmap_structure_find_item (structure, DMAP_CC_MUSR);
    if (item == NULL) {
        g_debug ("Could not find dmap.serverrevision item in /update");
        dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
        return;
    }

    priv->revision_number = g_value_get_int (&(item->content));
    dmap_connection_state_done (self, TRUE, DMAP_CONNECTION_ERROR_NONE);
}

static void
handle_database_info (DMAPConnection *self,
                      guint status,
                      GNode *structure,
                      gpointer user_data)
{
    DMAPConnectionPrivate *priv = self->priv;
    DMAPStructureItem *item = NULL;
    GNode *listing_node;
    gint n_databases = 0;

    /* get a list of databases, there should be only 1 */

    if (structure == NULL || SOUP_STATUS_IS_SUCCESSFUL (status) == FALSE) {
        dmap_connection_state_done (self, FALSE, dmap_connection_map_error (status));
        return;
    }

    item = dmap_structure_find_item (structure, DMAP_CC_MRCO);
    if (item == NULL) {
        g_debug ("Could not find dmap.returnedcount item in /databases");
        dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
        return;
    }

    n_databases = g_value_get_int (&(item->content));
    if (n_databases != 1) {
        g_debug ("Host seems to have more than 1 database, how strange\n");
    }

    listing_node = dmap_structure_find_node (structure, DMAP_CC_MLCL);
    if (listing_node == NULL) {
        g_debug ("Could not find dmap.listing item in /databases");
        dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
        return;
    }

    item = dmap_structure_find_item (listing_node->children, DMAP_CC_MIID);
    if (item == NULL) {
        g_debug ("Could not find dmap.itemid item in /databases");
        dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
        return;
    }

    priv->database_id = g_value_get_int (&(item->content));
    dmap_connection_state_done (self, TRUE, DMAP_CONNECTION_ERROR_NONE);
}

static void
handle_logout (DMAPConnection *self,
               guint status,
               GNode *structure,
               gpointer user_data)
{
    /* is there any point handling errors here? */
    dmap_connection_state_done (self, TRUE, DMAP_CONNECTION_ERROR_NONE);
    g_signal_emit (self, dmap_connection_signals [DISCONNECTED], 0);
}

static void
dmap_connection_state_done (DMAPConnection *self,
                            gboolean result,
                            DMAPConnectionError error)
{
    DMAPConnectionPrivate *priv = self->priv;

    if (result == FALSE) {
        priv->state = DMAP_DONE;
        priv->result = FALSE;
        g_signal_emit (self, dmap_connection_signals [ERROR], 0, error);
        return;
    } else {
        switch (priv->state) {
        case DMAP_GET_DB_INFO:
            priv->state = DMAP_CONNECTED;
            g_signal_emit (self, dmap_connection_signals [CONNECTED], 0);
            break;

        case DMAP_LOGOUT:
            priv->state = DMAP_DONE;
            break;

        case DMAP_DONE:
            /* uhh.. */
            g_debug ("This should never happen.");
            break;

        default:
            /* in most states, we just move on to the next */
            if (priv->state > DMAP_DONE) {
                g_debug ("This should REALLY never happen.");
                return;
            }
            priv->state++;
            break;
        }
    }

    dmap_connection_do_something (self);
}

static void
dmap_connection_do_something (DMAPConnection *self)
{
    DMAPConnectionPrivate *priv = self->priv;
    gchar *path;

    switch (priv->state) {
    case DMAP_GET_INFO:
        g_debug ("Getting server info");
        if (!http_get (self, "/server-info", FALSE, 0.0, 0, FALSE,
                   (DMAPResponseHandler) handle_server_info, NULL)) {
            g_debug ("Could not get self info");
            dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
        }
        break;

    case DMAP_GET_PASSWORD:
        g_debug ("Getting password");
        if (priv->password) {
            /* If the share went away while we were asking for the password,
             * DON't bother trying to log in.
             */
            if (priv->state != DMAP_GET_PASSWORD) {
                return;
            }
        }

        /* otherwise, fall through */
        priv->state = DMAP_LOGIN;

    case DMAP_LOGIN:
        g_debug ("Logging into server");
        if (!http_get (self, "/login", FALSE, 0.0, 0, FALSE,
                   (DMAPResponseHandler) handle_login, NULL)) {
            g_debug ("Could not login to server");
            /* FIXME: set state back to GET_PASSWORD to try again */
            dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
        }

        break;

    case DMAP_GET_REVISION_NUMBER:
        if (priv->need_revision_number) {
            g_debug ("Getting server database revision number");
            path = g_strdup_printf ("/update?session-id=%u&revision-number=1", priv->session_id);
            if (!http_get (self, path, TRUE, priv->protocol_version, 0, FALSE,
                       (DMAPResponseHandler) handle_update, NULL)) {
                g_debug ("Could not get server database revision number");
                dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
            }
            g_free (path);
            break;
        }

        priv->state = DMAP_GET_DB_INFO;

    case DMAP_GET_DB_INFO:
        g_debug ("Getting database info");
        if (priv->need_revision_number)
            path = g_strdup_printf ("/databases?session-id=%u&revision-number=%d",
                    priv->session_id, priv->revision_number);
        else
            path = g_strdup_printf ("/databases?session-id=%u",
                    priv->session_id);
        if (!http_get (self, path, TRUE, priv->protocol_version, 0, FALSE,
                   (DMAPResponseHandler) handle_database_info, NULL)) {
            g_debug ("Could not get database info");
            dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
        }
        g_free (path);
        break;

    case DMAP_LOGOUT:
        g_debug ("Logging out of server");
        path = g_strdup_printf ("/logout?session-id=%u", priv->session_id);
        if (!http_get (self, path, TRUE, priv->protocol_version, 0, FALSE,
                   (DMAPResponseHandler) handle_logout, NULL)) {
            g_debug ("Could not log out of server");
            dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_GENERIC);
        }

        g_free (path);
        break;

    default:
        break;
    }
}

static DMAPConnectionError
dmap_connection_map_error (guint status)
{
    DMAPConnectionError ret;
    switch (status) {
    case 403:
        ret = DMAP_CONNECTION_ERROR_FORBIDDEN;
        break;
    case 503:
        ret = DMAP_CONNECTION_ERROR_SERVICE_UNAVAILABLE;
        break;
    default:
        ret = DMAP_CONNECTION_ERROR_GENERIC;
        break;
    }
    return ret;
}

DMAPConnection *
dmap_connection_new (const gchar *name,
                     const gchar *host,
                     gint port,
                     const gchar *password)
{
    DMAPConnection *self = DMAP_CONNECTION (g_object_new (TYPE_DMAP_CONNECTION,
            "name", name,
            "host", host,
            "port", port,
            NULL));

    if (password) {
        g_object_set (self, "password", password, NULL);
    }

    return self;
}

gboolean
dmap_connection_connect (DMAPConnection *self)
{
    DMAPConnectionPrivate *priv;
    gchar *path = NULL;

    g_return_val_if_fail (self != NULL, FALSE);

    priv = self->priv;
    priv->result = TRUE;

    g_debug ("Creating new self to %s:%d", priv->host, priv->port);

    priv->session = soup_session_async_new ();
    path = g_strdup_printf ("http://%s:%d", priv->host, priv->port);
    priv->base_uri = soup_uri_new (path);
    g_free (path);

    if (priv->base_uri == NULL) {
        g_debug ("Error parsing http://%s:%d", priv->host, priv->port);
        return FALSE;
    }

    priv->state = DMAP_GET_INFO;
    dmap_connection_do_something (self);

    return TRUE;
}

void
dmap_connection_disconnect (DMAPConnection *self)
{
    DMAPConnectionPrivate *priv = self->priv;
    gint old_state = priv->state;

    if (priv->state >= DMAP_LOGOUT)
    {
        return;
    }

    priv->state = DMAP_LOGOUT;

    if (old_state <= DMAP_LOGIN) {
        /* we haven't logged in yet, so we don't need to do anything here */
        dmap_connection_state_done (self, FALSE, DMAP_CONNECTION_ERROR_NONE);
    } else {
        priv->result = TRUE;
        dmap_connection_do_something (self);
    }
}

gboolean
dmap_connection_get (DMAPConnection *self,
                     const gchar *path,
                     gboolean need_hash,
                     DMAPResponseHandler handler,
                     gpointer user_data)
{
    return http_get (self, path, need_hash,
            self->priv->protocol_version, 0, FALSE,
            (DMAPResponseHandler) handler, user_data);
}

gdouble
dmap_connection_get_protocol_version (DMAPConnection *self)
{
    g_return_val_if_fail (self != NULL, 0);
    return self->priv->protocol_version;
}

guint32
dmap_connection_get_session_id (DMAPConnection *self)
{
    g_return_val_if_fail (self != NULL, 0);
    return self->priv->session_id;
}

gint
dmap_connection_get_database_id (DMAPConnection *self)
{
    g_return_val_if_fail (self != NULL, -1);
    return self->priv->database_id;
}

gint
dmap_connection_get_revision_number (DMAPConnection *self)
{
    g_return_val_if_fail (self != NULL, -1);
    return self->priv->revision_number;
}

DMAPItem *
dmap_connection_get_root_item (DMAPConnection *self)
{
    g_return_val_if_fail (self != NULL, NULL);
    if (!self->priv->root_item) {
        DMAPConnectionClass *klass = DMAP_CONNECTION_GET_CLASS (self);
        self->priv->root_item = klass->create_root_item (self);
    }
    return g_object_ref (self->priv->root_item);
}
